@@ -41,7 +41,12 @@
 #if LIBAVCODEC_VERSION_INT >= AV_VERSION_INT(53, 34, 0)
 #include <libavutil/channel_layout.h>
 #include <libavutil/mem.h>
-#define USE_AVCODEC_FRAME
+#define USE_AVCODEC_FRAME 1
+#endif
+
+#if LIBAVCODEC_VERSION_INT >= AV_VERSION_INT(58, 91, 0)
+#include <libavcodec/packet.h>
+#define USE_AVCODEC_PACKET_ALLOC
 #endif
 
 #if LIBAVCODEC_VERSION_INT < AV_VERSION_INT(54, 0, 0)
@@ -62,6 +67,12 @@
 #define AV_CODEC_ID_AC3 CODEC_ID_AC3
 #endif
 
+#if LIBAVCODEC_VERSION_INT < AV_VERSION_INT(56, 56, 0)
+#ifndef AV_INPUT_BUFFER_PADDING_SIZE
+#define AV_INPUT_BUFFER_PADDING_SIZE   FF_INPUT_BUFFER_PADDING_SIZE
+#endif
+#endif
+
 #if LIBAVCODEC_VERSION_INT < 0x371c01
 #define av_frame_alloc avcodec_alloc_frame
 #define av_frame_free avcodec_free_frame
@@ -72,20 +83,29 @@ struct a52_ctx {
 	snd_pcm_t *slave;
 	AVCodec *codec;
 	AVCodecContext *avctx;
+	snd_pcm_format_t src_format;
+	unsigned int src_sample_bits;
+	unsigned int src_sample_bytes;
 	snd_pcm_format_t format;
 	int av_format;
 	unsigned int channels;
 	unsigned int rate;
 	unsigned int bitrate;
-	short *inbuf;
+	void *inbuf;
 	unsigned char *outbuf;
+	unsigned char *outbuf1;
+	unsigned char *outbuf2;
 	int outbuf_size;
-	snd_pcm_uframes_t transfer;
 	int remain;
 	int filled;
 	unsigned int slave_period_size;
 	unsigned int slave_buffer_size;
+	snd_pcm_uframes_t pointer;
+	snd_pcm_uframes_t boundary;
 	snd_pcm_hw_params_t *hw_params;
+#ifdef USE_AVCODEC_PACKET_ALLOC
+	AVPacket *pkt;
+#endif
 #ifdef USE_AVCODEC_FRAME
 	AVFrame *frame;
 	int is_planar;
@@ -98,71 +118,104 @@ struct a52_ctx {
 #define use_planar(rec)		0
 #endif
 
-#if LIBAVCODEC_VERSION_INT >= AV_VERSION_INT(53, 34, 0)
+#if LIBAVCODEC_VERSION_INT >= AV_VERSION_INT(58, 91, 0)
+static int do_encode(struct a52_ctx *rec)
+{
+	AVPacket *pkt = rec->pkt;
+	int ret;
+
+	ret = avcodec_send_frame(rec->avctx, rec->frame);
+	if (ret < 0)
+		return -EINVAL;
+	ret = avcodec_receive_packet(rec->avctx, pkt);
+	if (ret < 0)
+		return -EINVAL;
+
+	if (pkt->size > rec->outbuf_size - 8)
+		return -EINVAL;
+	memcpy(rec->outbuf1 + 8, pkt->data, pkt->size);
+
+	return pkt->size;
+}
+#elif LIBAVCODEC_VERSION_INT >= AV_VERSION_INT(53, 34, 0)
 static int do_encode(struct a52_ctx *rec)
 {
 	AVPacket pkt = {
-		.data = rec->outbuf + 8,
+		.data = rec->outbuf1 + 8,
 		.size = rec->outbuf_size - 8
 	};
-	int got_frame;
+	int ret, got_frame;
+
+	ret = avcodec_encode_audio2(rec->avctx, &pkt, rec->frame, &got_frame);
+	if (ret < 0)
+		return -EINVAL;
 
-	avcodec_encode_audio2(rec->avctx, &pkt, rec->frame, &got_frame);
 	return pkt.size;
 }
 #else
 static int do_encode(struct a52_ctx *rec)
 {
-	return avcodec_encode_audio(rec->avctx, rec->outbuf + 8,
-				    rec->outbuf_size - 8,
-				    rec->inbuf);
+	int ret = avcodec_encode_audio(rec->avctx, rec->outbuf1 + 8,
+				       rec->outbuf_size - 8,
+				       rec->inbuf);
+	if (ret < 0)
+		return -EINVAL;
+
+	return ret;
 }
 #endif
 
 /* convert the PCM data to A52 stream in IEC958 */
-static void convert_data(struct a52_ctx *rec)
+static int convert_data(struct a52_ctx *rec)
 {
+	unsigned char *buf;
 	int out_bytes = do_encode(rec);
 
-	rec->outbuf[0] = 0xf8; /* sync words */
-	rec->outbuf[1] = 0x72;
-	rec->outbuf[2] = 0x4e;
-	rec->outbuf[3] = 0x1f;
-	rec->outbuf[4] = rec->outbuf[13] & 7; /* bsmod */
-	rec->outbuf[5] = 0x01; /* data type */
-	rec->outbuf[6] = ((out_bytes * 8) >> 8) & 0xff;
-	rec->outbuf[7] = (out_bytes * 8) & 0xff;
+	if (out_bytes < 0)
+		return out_bytes;
+
+	buf = rec->outbuf1;
+	buf[0] = 0xf8; /* sync words */
+	buf[1] = 0x72;
+	buf[2] = 0x4e;
+	buf[3] = 0x1f;
+	buf[4] = buf[13] & 7; /* bsmod */
+	buf[5] = 0x01; /* data type */
+	buf[6] = ((out_bytes * 8) >> 8) & 0xff;
+	buf[7] = (out_bytes * 8) & 0xff;
 	/* swap bytes for little-endian 16bit */
-	if (rec->format == SND_PCM_FORMAT_S16_LE)
-		swab(rec->outbuf, rec->outbuf, out_bytes + 8);
-	memset(rec->outbuf +  8 + out_bytes, 0,
+	if (rec->format == SND_PCM_FORMAT_S16_LE) {
+		swab(rec->outbuf1, rec->outbuf2, out_bytes + 8);
+		buf = rec->outbuf2;
+	}
+	rec->outbuf = buf;
+	memset(buf + 8 + out_bytes, 0,
 	       rec->outbuf_size - 8 - out_bytes);
 	rec->remain = rec->outbuf_size / 4;
 	rec->filled = 0;
+
+	return 0;
 }
 
 /* write pending encoded data to the slave pcm */
 static int write_out_pending(snd_pcm_ioplug_t *io, struct a52_ctx *rec)
 {
-	int err, ofs = 0;
-
-	if (! rec->remain)
-		return 0;
+	snd_pcm_sframes_t ret;
+	unsigned int ofs;
 
 	while (rec->remain) {
-		err = snd_pcm_writei(rec->slave, rec->outbuf + ofs, rec->remain);
-		if (err < 0) {
-			if (err == -EPIPE)
+		ofs = (rec->avctx->frame_size - rec->remain) * 4;
+		ret = snd_pcm_writei(rec->slave, rec->outbuf + ofs, rec->remain);
+		if (ret < 0) {
+			if (ret == -EPIPE)
 				io->state = SND_PCM_STATE_XRUN;
-			return err;
-		} else if (! err)
+			if (ret == -EAGAIN)
+				break;
+			return ret;
+		} else if (! ret)
 			break;
-		if (err < rec->remain)
-			ofs += (rec->remain - err) * 4;
-		rec->remain -= err;
+		rec->remain -= ret;
 	}
-	if (rec->remain && ofs)
-		memmove(rec->outbuf, rec->outbuf + ofs, rec->remain * 4);
 	return 0;
 }
 
@@ -176,8 +229,8 @@ static void clear_remaining_planar_data(snd_pcm_ioplug_t *io)
 	unsigned int i;
 
 	for (i = 0; i < io->channels; i++)
-		memset(rec->frame->data[i] + rec->filled * 2, 0,
-		       (rec->avctx->frame_size - rec->filled) * 2);
+		memset(rec->frame->data[i] + rec->filled * rec->src_sample_bytes, 0,
+		       (rec->avctx->frame_size - rec->filled) * rec->src_sample_bytes);
 }
 #else
 #define clear_remaining_planar_data(io) /*NOP*/
@@ -195,10 +248,12 @@ static int a52_drain(snd_pcm_ioplug_t *io)
 		if (use_planar(rec))
 			clear_remaining_planar_data(io);
 		else {
-			memset(rec->inbuf + rec->filled * io->channels, 0,
-			       (rec->avctx->frame_size - rec->filled) * io->channels * 2);
+			memset(rec->inbuf + rec->filled * io->channels * rec->src_sample_bytes, 0,
+			       (rec->avctx->frame_size - rec->filled) * io->channels * rec->src_sample_bytes);
 		}
-		convert_data(rec);
+		err = convert_data(rec);
+		if (err < 0)
+			return err;
 	}
 	err = write_out_pending(io, rec);
 	if (err < 0)
@@ -208,7 +263,8 @@ static int a52_drain(snd_pcm_ioplug_t *io)
 }
 
 /* check whether the areas consist of a continuous interleaved stream */
-static int check_interleaved(const snd_pcm_channel_area_t *areas,
+static int check_interleaved(struct a52_ctx *rec,
+			     const snd_pcm_channel_area_t *areas,
 			     unsigned int channels)
 {
 	unsigned int ch;
@@ -218,8 +274,8 @@ static int check_interleaved(const snd_pcm_channel_area_t *areas,
 
 	for (ch = 0; ch < channels; ch++) {
 		if (areas[ch].addr != areas[0].addr ||
-		    areas[ch].first != ch * 16 ||
-		    areas[ch].step != channels * 16)
+		    areas[ch].first != ch * rec->src_sample_bits ||
+		    areas[ch].step != channels * rec->src_sample_bits)
 			return 0;
 	}
 	return 1;
@@ -237,8 +293,7 @@ static int fill_data(snd_pcm_ioplug_t *io,
 {
 	struct a52_ctx *rec = io->private_data;
 	unsigned int len = rec->avctx->frame_size - rec->filled;
-	short *src, *dst;
-	unsigned int src_step;
+	void *_dst;
 	int err;
 	static unsigned int ch_index[3][6] = {
 		{ 0, 1 },
@@ -255,16 +310,25 @@ static int fill_data(snd_pcm_ioplug_t *io,
 	if ((err = write_out_pending(io, rec)) < 0)
 		return err;
 
+	/* If there are still frames left in outbuf, we can't
+	 * accept a full a52 frame, because this would overwrite
+	 * the frames in outbuf. This should not happen! The a52_pointer()
+	 * callback should limit the transferred frames correctly. */
+	if (rec->remain && len) {
+		SNDERR("fill data issue (remain is %i)", rec->remain);
+		len--;
+	}
+
 	if (size > len)
 		size = len;
 
-	dst = rec->inbuf + rec->filled * io->channels;
+	_dst = rec->inbuf + rec->filled * io->channels * rec->src_sample_bytes;
 	if (!use_planar(rec) && interleaved) {
-		memcpy(dst, areas->addr + offset * io->channels * 2,
-		       size * io->channels * 2);
-	} else {
-		unsigned int i, ch, dst_step;
-		short *dst1;
+		memcpy(_dst, areas->addr + offset * io->channels * rec->src_sample_bytes,
+		       size * io->channels * rec->src_sample_bytes);
+	} else if (rec->src_sample_bits == 16) {
+		unsigned int i, ch, src_step, dst_step;
+		short *src, *dst = _dst, *dst1;
 
 		/* flatten copy to n-channel interleaved */
 		dst_step = io->channels;
@@ -275,8 +339,8 @@ static int fill_data(snd_pcm_ioplug_t *io,
 					(ap->first + offset * ap->step) / 8);
 
 #ifdef USE_AVCODEC_FRAME
-			if (use_planar(rec)) {
-				memcpy(rec->frame->data[ch], src, size * 2);
+			if (use_planar(rec) && !interleaved) {
+				memcpy(rec->frame->data[ch] + rec->filled * 2, src, size * 2);
 				continue;
 			}
 #endif
@@ -288,10 +352,40 @@ static int fill_data(snd_pcm_ioplug_t *io,
 				dst1 += dst_step;
 			}
 		}
+	} else if (rec->src_sample_bits == 32) {
+		unsigned int i, ch, src_step, dst_step;
+		int *src, *dst = _dst, *dst1;
+
+		/* flatten copy to n-channel interleaved */
+		dst_step = io->channels;
+		for (ch = 0; ch < io->channels; ch++, dst++) {
+			const snd_pcm_channel_area_t *ap;
+			ap = &areas[ch_index[io->channels / 2 - 1][ch]];
+			src = (int *)(ap->addr +
+					(ap->first + offset * ap->step) / 8);
+
+#ifdef USE_AVCODEC_FRAME
+			if (use_planar(rec) && !interleaved) {
+				memcpy(rec->frame->data[ch] + rec->filled * 4, src, size * 4);
+				continue;
+			}
+#endif
+			dst1 = dst;
+			src_step = ap->step / 32; /* in word */
+			for (i = 0; i < size; i++) {
+				*dst1 = *src;
+				src += src_step;
+				dst1 += dst_step;
+			}
+		}
+	} else {
+		return -EIO;
 	}
 	rec->filled += size;
 	if (rec->filled == rec->avctx->frame_size) {
-		convert_data(rec);
+		err = convert_data(rec);
+		if (err < 0)
+			return err;
 		write_out_pending(io, rec);
 	}
 	return (int)size;
@@ -308,16 +402,17 @@ static snd_pcm_sframes_t a52_transfer(snd_pcm_ioplug_t *io,
 	struct a52_ctx *rec = io->private_data;
 	snd_pcm_sframes_t result = 0;
 	int err = 0;
-	int interleaved = check_interleaved(areas, io->channels);
+	int interleaved = check_interleaved(rec, areas, io->channels);
 
 	do {
 		err = fill_data(io, areas, offset, size, interleaved);
-		if (err < 0)
+		if (err <= 0)
 			break;
 		offset += (unsigned int)err;
 		size -= (unsigned int)err;
 		result += err;
-		rec->transfer += err;
+		rec->pointer += err;
+		rec->pointer %= rec->boundary;
 	} while (size);
 	return result > 0 ? result : err;
 }
@@ -325,39 +420,45 @@ static snd_pcm_sframes_t a52_transfer(snd_pcm_ioplug_t *io,
 /*
  * pointer callback
  *
- * Calculate the current position from the delay of slave PCM
+ * Calculate the current position from the available frames of slave PCM
  */
 static snd_pcm_sframes_t a52_pointer(snd_pcm_ioplug_t *io)
 {
 	struct a52_ctx *rec = io->private_data;
-	snd_pcm_sframes_t delay;
+	snd_pcm_sframes_t avail, delay;
 	snd_pcm_state_t state;
-	int err;
 
 	state = snd_pcm_state(rec->slave);
 	switch (state) {
 	case SND_PCM_STATE_RUNNING:
 	case SND_PCM_STATE_DRAINING:
-		if ((err = snd_pcm_delay(rec->slave, &delay)) < 0)
-			return err;
 		break;
 	case SND_PCM_STATE_XRUN:
-	case SND_PCM_STATE_SUSPENDED:
 		return -EPIPE;
+	case SND_PCM_STATE_SUSPENDED:
+		return -ESTRPIPE;
 	default:
 		return 0;
 	}
 
-	if (delay < 0 || delay >= (snd_pcm_sframes_t)rec->slave_buffer_size)
-		delay = 0;
-	delay = (snd_pcm_sframes_t)io->appl_ptr - delay;
-	if (delay < 0) {
-		delay += io->buffer_size;
-		if (delay < 0)
-			delay = 0;
-	}
-	delay %= io->buffer_size;
-	return delay;
+	/* Write what we have from outbuf. */
+	write_out_pending(io, rec);
+
+	avail = snd_pcm_avail(rec->slave);
+	if (avail < 0)
+		return avail;
+
+	/* get buffer delay without additional FIFO information */
+	delay = rec->slave_buffer_size - avail;
+	while (delay < 0)
+		delay += rec->slave_buffer_size;
+
+	avail = rec->pointer - delay - rec->remain - rec->filled;
+#ifdef SND_PCM_IOPLUG_FLAG_BOUNDARY_WA
+	return avail % rec->boundary;
+#else
+	return avail % io->buffer_size;
+#endif
 }
 
 /* set up the fixed parameters of slave PCM hw_parmas */
@@ -450,6 +551,25 @@ static int a52_hw_free(snd_pcm_ioplug_t *io)
 	return snd_pcm_hw_free(rec->slave);
 }
 
+/*
+ * dump callback
+ */
+static void a52_dump(snd_pcm_ioplug_t *io, snd_output_t *out)
+{
+	struct a52_ctx *rec = io->private_data;
+	snd_pcm_t *pcm = io->pcm;
+
+	snd_output_printf(out, "%s\n", io->name);
+	snd_output_printf(out, "Its setup is:\n");
+	snd_pcm_dump_setup(pcm, out);
+	snd_output_printf(out, "  %-13s: %s\n", "av_format", av_get_sample_fmt_name(rec->av_format));
+	snd_output_printf(out, "  %-13s: %i\n", "av_frame_size", rec->avctx ? rec->avctx->frame_size : -1);
+	snd_output_printf(out, "  %-13s: %i\n", "remain", rec->remain);
+	snd_output_printf(out, "  %-13s: %i\n", "filled", rec->filled);
+	snd_output_printf(out, "Slave: ");
+	snd_pcm_dump(rec->slave, out);
+}
+
 /*
  * sw_params callback
  *
@@ -464,6 +584,7 @@ static int a52_sw_params(snd_pcm_ioplug_t *io, snd_pcm_sw_params_t *params)
 
 	snd_pcm_sw_params_get_avail_min(params, &avail_min);
 	snd_pcm_sw_params_get_start_threshold(params, &start_threshold);
+	snd_pcm_sw_params_get_boundary(params, &rec->boundary);
 
 	len = avail_min;
 	len += (int)rec->slave_buffer_size - (int)io->buffer_size;
@@ -513,21 +634,27 @@ static void a52_free(struct a52_ctx *rec)
 	}
 
 #ifdef USE_AVCODEC_FRAME
-	if (rec->frame) {
+#if LIBAVCODEC_VERSION_INT < AV_VERSION_INT(58, 91, 0)
+	if (rec->frame)
 		av_freep(&rec->frame->data[0]);
-		rec->inbuf = NULL;
-	}
+#endif
 #if LIBAVCODEC_VERSION_INT >= AV_VERSION_INT(54, 28, 0)
 	av_frame_free(&rec->frame);
 #else
 	av_freep(&rec->frame);
 #endif
-#endif
-
+#else /* USE_AVCODEC_FRAME */
 	free(rec->inbuf);
 	rec->inbuf = NULL;
-	free(rec->outbuf);
-	rec->outbuf = NULL;
+#endif /* USE_AVCODEC_FRAME */
+
+#ifdef USE_AVCODEC_PACKET_ALLOC
+	av_packet_free(&rec->pkt);
+#endif
+	free(rec->outbuf2);
+	rec->outbuf2 = NULL;
+	free(rec->outbuf1);
+	rec->outbuf1 = NULL;
 }
 
 /*
@@ -565,14 +692,22 @@ static int alloc_input_buffer(snd_pcm_ioplug_t *io)
 	rec->frame = av_frame_alloc();
 	if (!rec->frame)
 		return -ENOMEM;
+	rec->frame->nb_samples = rec->avctx->frame_size;
+#if LIBAVCODEC_VERSION_INT >= AV_VERSION_INT(58, 91, 0)
+	rec->frame->format = rec->avctx->sample_fmt;
+	rec->frame->channels = rec->avctx->channels;
+	rec->frame->channel_layout = rec->avctx->channel_layout;
+	if (av_frame_get_buffer(rec->frame, 0))
+		return -ENOMEM;
+#else
 	if (av_samples_alloc(rec->frame->data, rec->frame->linesize,
 			     io->channels, rec->avctx->frame_size,
 			     rec->avctx->sample_fmt, 0) < 0)
 		return -ENOMEM;
-	rec->frame->nb_samples = rec->avctx->frame_size;
-	rec->inbuf = (short *)rec->frame->data[0];
+#endif
+	rec->inbuf = rec->frame->data[0];
 #else
-	rec->inbuf = malloc(rec->avctx->frame_size * 2 * io->channels);
+	rec->inbuf = malloc(rec->avctx->frame_size * io->channels * rec->src_sample_bytes);
 #endif
 	if (!rec->inbuf)
 		return -ENOMEM;
@@ -610,15 +745,28 @@ static int a52_prepare(snd_pcm_ioplug_t *io)
 	if (err < 0)
 		return -EINVAL;
 
-	rec->outbuf_size = rec->avctx->frame_size * 4;
-	rec->outbuf = malloc(rec->outbuf_size);
-	if (! rec->outbuf)
+#ifdef USE_AVCODEC_PACKET_ALLOC
+	rec->pkt = av_packet_alloc();
+	if (!rec->pkt)
 		return -ENOMEM;
+#endif
+
+	rec->outbuf_size = rec->avctx->frame_size * 4;
+	rec->outbuf1 = malloc(rec->outbuf_size + AV_INPUT_BUFFER_PADDING_SIZE);
+	if (! rec->outbuf1)
+		return -ENOMEM;
+	memset(rec->outbuf1 + rec->outbuf_size, 0, AV_INPUT_BUFFER_PADDING_SIZE);
+
+	if (rec->format == SND_PCM_FORMAT_S16_LE) {
+		rec->outbuf2 = malloc(rec->outbuf_size);
+		if ( !rec->outbuf2)
+			return -ENOMEM;
+	}
 
 	if (alloc_input_buffer(io))
 		return -ENOMEM;
 
-	rec->transfer = 0;
+	rec->pointer = 0;
 	rec->remain = 0;
 	rec->filled = 0;
 
@@ -654,10 +802,12 @@ static int a52_poll_revents(snd_pcm_ioplug_t *io, struct pollfd *pfd,
 static int a52_close(snd_pcm_ioplug_t *io)
 {
 	struct a52_ctx *rec = io->private_data;
+	snd_pcm_t *slave = rec->slave;
 
 	a52_free(rec);
-	if (rec->slave)
-		return snd_pcm_close(rec->slave);
+	free(rec);
+	if (slave)
+		return snd_pcm_close(slave);
 	return 0;
 }
 			      
@@ -722,6 +872,7 @@ static snd_pcm_ioplug_callback_t a52_ops = {
 	.close = a52_close,
 	.hw_params = a52_hw_params,
 	.hw_free = a52_hw_free,
+	.dump = a52_dump,
 	.sw_params = a52_sw_params,
 	.prepare = a52_prepare,
 	.drain = a52_drain,
@@ -758,8 +909,20 @@ static int a52_set_hw_constraint(struct a52_ctx *rec)
 		SND_PCM_ACCESS_MMAP_NONINTERLEAVED,
 		SND_PCM_ACCESS_RW_NONINTERLEAVED
 	};
-	unsigned int formats[] = { SND_PCM_FORMAT_S16 };
-	int err;
+	static struct format {
+		int av;
+		snd_pcm_format_t alib;
+	} formats[] = {
+		{ .av = AV_SAMPLE_FMT_S16, .alib = SND_PCM_FORMAT_S16 },
+#if LIBAVCODEC_VERSION_INT >= AV_VERSION_INT(52, 95, 0) && USE_AVCODEC_FRAME
+		{ .av = AV_SAMPLE_FMT_S16P, .alib = SND_PCM_FORMAT_S16 },
+		{ .av = AV_SAMPLE_FMT_S32, .alib = SND_PCM_FORMAT_S32 },
+		{ .av = AV_SAMPLE_FMT_S32P, .alib = SND_PCM_FORMAT_S32 },
+		{ .av = AV_SAMPLE_FMT_FLTP, .alib = SND_PCM_FORMAT_FLOAT }
+#endif
+	};
+	int err, dir;
+	unsigned int i, fmt;
 	snd_pcm_uframes_t buffer_max;
 	unsigned int period_bytes, max_periods;
 
@@ -776,8 +939,22 @@ static int a52_set_hw_constraint(struct a52_ctx *rec)
 	if (err < 0)
 		return err;
 
+	rec->src_format = SND_PCM_FORMAT_UNKNOWN;
+	for (i = 0; i < ARRAY_SIZE(formats); i++)
+		if (formats[i].av == rec->av_format) {
+			rec->src_format = formats[i].alib;
+			break;
+		}
+	if (rec->src_format == SND_PCM_FORMAT_UNKNOWN) {
+		SNDERR("A/V format '%s' is not supported", av_get_sample_fmt_name(rec->av_format));
+		return -EINVAL;
+	}
+	fmt = rec->src_format;
+	rec->src_sample_bits = snd_pcm_format_physical_width(rec->src_format);
+	rec->src_sample_bytes = rec->src_sample_bits / 8;
+
 	if ((err = snd_pcm_ioplug_set_param_list(&rec->io, SND_PCM_IOPLUG_HW_FORMAT,
-						 ARRAY_SIZE(formats), formats)) < 0 ||
+						 1, &fmt)) < 0 ||
 	    (err = snd_pcm_ioplug_set_param_minmax(&rec->io, SND_PCM_IOPLUG_HW_CHANNELS,
 						   rec->channels, rec->channels)) < 0 ||
 	    (err = snd_pcm_ioplug_set_param_minmax(&rec->io, SND_PCM_IOPLUG_HW_RATE,
@@ -788,8 +965,11 @@ static int a52_set_hw_constraint(struct a52_ctx *rec)
 		return err;
 
 	snd_pcm_hw_params_get_buffer_size_max(rec->hw_params, &buffer_max);
+	dir = -1;
+	snd_pcm_hw_params_get_periods_max(rec->hw_params, &max_periods, &dir);
 	period_bytes = A52_FRAME_SIZE * 2 * rec->channels;
-	max_periods = buffer_max / A52_FRAME_SIZE;
+	if (buffer_max / A52_FRAME_SIZE < max_periods)
+		max_periods = buffer_max / A52_FRAME_SIZE;
 
 	if ((err = snd_pcm_ioplug_set_param_minmax(&rec->io, SND_PCM_IOPLUG_HW_PERIOD_BYTES,
 						   period_bytes, period_bytes)) < 0 ||
@@ -809,11 +989,12 @@ SND_PCM_PLUGIN_DEFINE_FUNC(a52)
 	int err;
 	const char *card = NULL;
 	const char *pcm_string = NULL;
+	const char *avcodec = NULL;
 	unsigned int rate = 48000;
 	unsigned int bitrate = 448;
 	unsigned int channels = 6;
 	snd_pcm_format_t format = SND_PCM_FORMAT_S16_LE;
-	char devstr[128], tmpcard[8];
+	char devstr[128], tmpcard[16];
 	struct a52_ctx *rec;
 	
 	if (stream != SND_PCM_STREAM_PLAYBACK) {
@@ -906,6 +1087,16 @@ SND_PCM_PLUGIN_DEFINE_FUNC(a52)
 			}
 			continue;
 		}
+		if (strcmp(id, "avcodec") == 0) {
+			const char *str;
+			err = snd_config_get_string(n, &str);
+			if (err < 0) {
+				SNDERR("invalid type for %s", id);
+				return -EINVAL;
+			}
+			avcodec = str;
+			continue;
+		}
 		SNDERR("Unknown field %s", id);
 		return -EINVAL;
 	}
@@ -916,6 +1107,7 @@ SND_PCM_PLUGIN_DEFINE_FUNC(a52)
 		return -ENOMEM;
 	}
 
+	rec->src_format = SND_PCM_FORMAT_UNKNOWN;
 	rec->rate = rate;
 	rec->bitrate = bitrate;
 	rec->channels = channels;
@@ -924,11 +1116,17 @@ SND_PCM_PLUGIN_DEFINE_FUNC(a52)
 #ifndef USE_AVCODEC_FRAME
 	avcodec_init();
 #endif
+#if LIBAVCODEC_VERSION_INT < AV_VERSION_INT(58, 91, 0)
 	avcodec_register_all();
+#endif
 
-	rec->codec = avcodec_find_encoder_by_name("ac3_fixed");
-	if (rec->codec == NULL)
-		rec->codec = avcodec_find_encoder_by_name("ac3");
+	if (avcodec) {
+		rec->codec = avcodec_find_encoder_by_name(avcodec);
+	} else {
+		rec->codec = avcodec_find_encoder_by_name("ac3_fixed");
+		if (rec->codec == NULL)
+			rec->codec = avcodec_find_encoder_by_name("ac3");
+	}
 	if (rec->codec == NULL) 
 		rec->codec = avcodec_find_encoder(AV_CODEC_ID_AC3);
 	if (rec->codec == NULL) {
@@ -939,7 +1137,7 @@ SND_PCM_PLUGIN_DEFINE_FUNC(a52)
 
 	if (! pcm_string || pcm_string[0] == '\0') {
 		snprintf(devstr, sizeof(devstr),
-			 "iec958:{AES0 0x%x AES1 0x%x AES2 0x%x AES3 0x%x %s%s}",
+			 "iec958:{AES0 0x%x AES1 0x%x AES2 0x%x AES3 0x%x%s%s}",
 			 IEC958_AES0_CON_EMPHASIS_NONE | IEC958_AES0_NONAUDIO |
 			 IEC958_AES0_CON_NOT_COPYRIGHT,
 			 IEC958_AES1_CON_ORIGINAL | IEC958_AES1_CON_PCM_CODER,
@@ -965,6 +1163,9 @@ SND_PCM_PLUGIN_DEFINE_FUNC(a52)
 	rec->io.mmap_rw = 0;
 	rec->io.callback = &a52_ops;
 	rec->io.private_data = rec;
+#ifdef SND_PCM_IOPLUG_FLAG_BOUNDARY_WA
+	rec->io.flags = SND_PCM_IOPLUG_FLAG_BOUNDARY_WA;
+#endif
 #ifdef USE_AVCODEC_FRAME
 	rec->av_format = rec->codec->sample_fmts[0];
 	rec->is_planar = av_sample_fmt_is_planar(rec->av_format);
@@ -978,7 +1179,7 @@ SND_PCM_PLUGIN_DEFINE_FUNC(a52)
 
 	if ((err = a52_set_hw_constraint(rec)) < 0) {
 		snd_pcm_ioplug_delete(&rec->io);
-		goto error;
+		return err;
 	}
 
 	*pcmp = rec->io.pcm;
