@@ -20,7 +20,9 @@
  *
  */
 
+#define _GNU_SOURCE
 #include <stdbool.h>
+#include <errno.h>
 #include <byteswap.h>
 #include <sys/shm.h>
 #include <sys/types.h>
@@ -28,6 +30,7 @@
 #include <jack/jack.h>
 #include <alsa/asoundlib.h>
 #include <alsa/pcm_external.h>
+#include <pthread.h>
 
 #define MAX_PERIODS_MULTIPLE 64
 
@@ -44,6 +47,8 @@ typedef struct {
 
 	int fd;
 	int activated;		/* jack is activated? */
+	pthread_mutex_t running_mutex;
+	int running;		/* jack is running? */
 
 	snd_pcm_jack_port_list_t **port_names;
 	unsigned int num_ports;
@@ -51,6 +56,7 @@ typedef struct {
 	snd_pcm_uframes_t hw_ptr;
 	unsigned int sample_bits;
 	snd_pcm_uframes_t min_avail;
+	int use_period_alignment;
 
 	snd_pcm_channel_area_t *areas;
 
@@ -158,6 +164,7 @@ static void snd_pcm_jack_free(snd_pcm_jack_t *jack)
 		free(jack->port_names);
 		jack->port_names = NULL;
 	}
+	pthread_mutex_destroy (&jack->running_mutex);
 	if (jack->fd >= 0)
 		close(jack->fd);
 	if (jack->io.poll_fd >= 0)
@@ -206,6 +213,18 @@ snd_pcm_jack_process_cb(jack_nframes_t nframes, snd_pcm_ioplug_t *io)
 	snd_pcm_jack_t *jack = io->private_data;
 	snd_pcm_uframes_t xfer = 0;
 	unsigned int channel;
+
+	if (pthread_mutex_trylock (&jack->running_mutex) == EBUSY) {
+		/* Note that locking should only ever fail if
+		 * snd_pcm_jack_start or snd_pcm_jack_stop is called at the
+		 * same time, in which case dropping the current buffer is not
+		 * an issue. */
+		return 0;
+	}
+	if (!jack->running) {
+		pthread_mutex_unlock (&jack->running_mutex);
+		return 0;
+	}
 	
 	for (channel = 0; channel < io->channels; channel++) {
 		jack->areas[channel].addr = 
@@ -267,6 +286,8 @@ snd_pcm_jack_process_cb(jack_nframes_t nframes, snd_pcm_ioplug_t *io)
 
 	pcm_poll_unblock_check(io); /* unblock socket for polling if needed */
 
+	pthread_mutex_unlock (&jack->running_mutex);
+
 	return 0;
 }
 
@@ -364,7 +385,10 @@ static int snd_pcm_jack_prepare(snd_pcm_ioplug_t *io)
 
 static int snd_pcm_jack_start(snd_pcm_ioplug_t *io)
 {
-	(void)io;
+	snd_pcm_jack_t *jack = io->private_data;
+	pthread_mutex_lock (&jack->running_mutex);
+	jack->running = 1;
+	pthread_mutex_unlock (&jack->running_mutex);
 	/*
 	 * Since the processing of jack_activate() and jack_connect() take a
 	 * while longer, snd_pcm_jack_start() was blocked.
@@ -381,7 +405,10 @@ static int snd_pcm_jack_start(snd_pcm_ioplug_t *io)
 
 static int snd_pcm_jack_stop(snd_pcm_ioplug_t *io)
 {
-	(void)io;
+	snd_pcm_jack_t *jack = io->private_data;
+	pthread_mutex_lock (&jack->running_mutex);
+	jack->running = 0;
+	pthread_mutex_unlock (&jack->running_mutex);
 	return 0;
 }
 
@@ -405,6 +432,13 @@ static int snd_pcm_jack_hw_free(snd_pcm_ioplug_t *io)
 	return 0;
 }
 
+static int snd_pcm_jack_sw_params(snd_pcm_ioplug_t *io, snd_pcm_sw_params_t *params)
+{
+	snd_pcm_jack_t *jack = io->private_data;
+	snd_pcm_sw_params_get_avail_min(params, &jack->min_avail);
+	return 0;
+}
+
 static snd_pcm_ioplug_callback_t jack_pcm_callback = {
 	.close = snd_pcm_jack_close,
 	.start = snd_pcm_jack_start,
@@ -413,6 +447,7 @@ static snd_pcm_ioplug_callback_t jack_pcm_callback = {
 	.hw_free = snd_pcm_jack_hw_free,
 	.prepare = snd_pcm_jack_prepare,
 	.poll_revents = snd_pcm_jack_poll_revents,
+	.sw_params = snd_pcm_jack_sw_params,
 };
 
 #define ARRAY_SIZE(ary)	(sizeof(ary)/sizeof(ary[0]))
@@ -450,8 +485,9 @@ static int jack_set_hw_constraint(snd_pcm_jack_t *jack)
 						   jack->num_ports, jack->num_ports)) < 0 ||
 	    (err = snd_pcm_ioplug_set_param_minmax(&jack->io, SND_PCM_IOPLUG_HW_RATE,
 						   rate, rate)) < 0 ||
-	    (err = snd_pcm_ioplug_set_param_list(&jack->io, SND_PCM_IOPLUG_HW_PERIOD_BYTES,
-						 ARRAY_SIZE(psize_list), psize_list)) < 0 ||
+	    (err = jack->use_period_alignment ?
+				snd_pcm_ioplug_set_param_list(&jack->io, SND_PCM_IOPLUG_HW_PERIOD_BYTES, ARRAY_SIZE(psize_list), psize_list) :
+				snd_pcm_ioplug_set_param_minmax(&jack->io, SND_PCM_IOPLUG_HW_PERIOD_BYTES, 128, 64*1024) ) < 0 ||
 	    (err = snd_pcm_ioplug_set_param_minmax(&jack->io, SND_PCM_IOPLUG_HW_PERIODS,
 						   2, 64)) < 0)
 		return err;
@@ -532,6 +568,7 @@ static int snd_pcm_jack_open(snd_pcm_t **pcmp, const char *name,
 			     const char *client_name,
 			     snd_config_t *playback_conf,
 			     snd_config_t *capture_conf,
+			     int use_period_alignment,
 			     snd_pcm_stream_t stream, int mode)
 {
 	snd_pcm_jack_t *jack;
@@ -545,8 +582,11 @@ static int snd_pcm_jack_open(snd_pcm_t **pcmp, const char *name,
 	if (!jack)
 		return -ENOMEM;
 
+	pthread_mutex_init (&jack->running_mutex, NULL);
+
 	jack->fd = -1;
 	jack->io.poll_fd = -1;
+	jack->use_period_alignment = use_period_alignment;
 
 	err = parse_ports(jack, stream == SND_PCM_STREAM_PLAYBACK ?
 			  playback_conf : capture_conf);
@@ -562,12 +602,16 @@ static int snd_pcm_jack_open(snd_pcm_t **pcmp, const char *name,
 		return -EINVAL;
 	}
 
-	if (client_name == NULL)
+	if (client_name == NULL) {
+		const char *pname = program_invocation_short_name;
+		if (!pname[0]) {
+			pname = "alsa-jack";
+		}
 		err = snprintf(jack_client_name, sizeof(jack_client_name),
-			       "alsa-jack.%s%s.%d.%d", name,
+			       "%s.%s.%d.%d", pname,
 			       stream == SND_PCM_STREAM_PLAYBACK ? "P" : "C",
 			       getpid(), num++);
-	else
+	} else
 		err = snprintf(jack_client_name, sizeof(jack_client_name),
 			       "%s", client_name);
 
@@ -635,6 +679,7 @@ SND_PCM_PLUGIN_DEFINE_FUNC(jack)
 	snd_config_t *capture_conf = NULL;
 	const char *client_name = NULL;
 	int err;
+	int align_jack_period = 1; /*by default we allow only JACK aligned period size*/
 	
 	snd_config_for_each(i, next, conf) {
 		snd_config_t *n = snd_config_iterator_entry(i);
@@ -663,11 +708,18 @@ SND_PCM_PLUGIN_DEFINE_FUNC(jack)
 			capture_conf = n;
 			continue;
 		}
+		if (strcmp(id, "align_psize") == 0) {
+			err = snd_config_get_bool(n);
+			if (err < 0)
+				return err;
+			align_jack_period = err ? 1 : 0;
+			continue;
+		}
 		SNDERR("Unknown field %s", id);
 		return -EINVAL;
 	}
 
-	err = snd_pcm_jack_open(pcmp, name, client_name, playback_conf, capture_conf, stream, mode);
+	err = snd_pcm_jack_open(pcmp, name, client_name, playback_conf, capture_conf, align_jack_period, stream, mode);
 
 	return err;
 }
